/**
 * Copyright 2022 Google LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *      https://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

package com.google.cast.sample.casthelloaudioandroid;


import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Color;
import android.media.AudioAttributes;
import android.media.AudioFocusRequest;
import android.media.AudioManager;
import android.media.MediaPlayer;
import android.media.session.MediaSession;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.support.v4.media.MediaBrowserCompat;
import android.support.v4.media.MediaDescriptionCompat;
import android.support.v4.media.MediaMetadataCompat;
import android.support.v4.media.session.MediaControllerCompat;
import android.support.v4.media.session.MediaSessionCompat;
import android.support.v4.media.session.PlaybackStateCompat;
import android.util.Log;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.media.MediaBrowserServiceCompat;
import androidx.media.session.MediaButtonReceiver;

import com.google.android.gms.cast.CastStatusCodes;
import com.google.android.gms.cast.MediaInfo;
import com.google.android.gms.cast.MediaLoadRequestData;
import com.google.android.gms.cast.MediaMetadata;
import com.google.android.gms.cast.SessionState;
import com.google.android.gms.cast.framework.CastContext;
import com.google.android.gms.cast.framework.CastSession;
import com.google.android.gms.cast.framework.SessionManagerListener;
import com.google.android.gms.cast.framework.SessionTransferCallback;
import com.google.android.gms.cast.framework.media.RemoteMediaClient;
import com.google.android.gms.common.images.WebImage;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

public class AudioPlayerService extends MediaBrowserServiceCompat implements AudioManager.OnAudioFocusChangeListener {

    private static final String LOG_TAG = AudioPlayerService.class.getSimpleName();
    private static final String MEDIA_ROOT_ID = "root";
    private static final String CHANNEL_ID = "com.google.cast.sample.casthelloaudioandroid.channel";
    private static final int NOTIFICATION_ID = 1011;

    /** Populate Metadata below before launch */
    private static final String KEY_MEDIA_ID = "Song1";
    private static final String KEY_MEDIA_TITLE = "Media Title"; //TODO - Add Media Title
    private static final String KEY_MEDIA_SUBTITLE = "Media Subtitle"; //TODO - Add Media Subtitle text
    private static final long KEY_MEDIA_DURATION = 0; //TODO - Add duration in Seconds
    private static final String KEY_MEDIA_ARTIST = "Artist Name"; //TODO - Add Artist Name
    private static final String KEY_MEDIA_ALBUM = "Album Name"; //TODO - Add Album Name
    private static final String KEY_MEDIA_URI = null; //TODO - Add URI for Media
    public static final String KEY_IMAGE_URI = null; //TODO - Add URI for Album ART

    private final int seekIncrement = 30*1000; // To be used with +/- skip30 implemented

    private NotificationManager notificationManager;

    private List<MediaBrowserCompat.MediaItem> mediaItems = new ArrayList<>();

    private MediaSessionCompat mediaSessionCompat;
    private MediaSessionCallback mediaSessionCallback;
    private PlaybackStateCompat.Builder playbackStateBuilder;
    private MediaMetadataCompat mediaMetadata;

    private MediaPlayer mediaPlayer;
    private static long currentPos = 0;

    private CastSession castSession;
    private CastContext castContext;
    private RemoteMediaClient remoteMediaClient;
    private CastSessionManagerListener castSessionManagerListener;
    private CastSessionTransferCallback castSessionTransferCallback;

    private AudioManager audioManager;
    private AudioAttributes audioAttributes;
    private AudioFocusRequest audioFocusRequest;
    private int focusResult;
    private final Object focusLock = new Object();

    /** Currently not/partially in use, WIP */
    private enum PlaybackLocation {
        LOCAL,
        REMOTE
    };
    private PlaybackLocation playbackLocation;

    @Override
    public void onCreate() {
        super.onCreate();
        Log.d(LOG_TAG, "onCreate called");
        checkMetadata();
        initMediaSession();

        castContext = CastContext.getSharedInstance();
        castSessionManagerListener = new CastSessionManagerListener();
        castContext.getSessionManager().addSessionManagerListener(castSessionManagerListener, CastSession.class);
        castSessionTransferCallback = new CastSessionTransferCallback();
        castContext.addSessionTransferCallback(castSessionTransferCallback);

        audioManager = (AudioManager) AudioPlayerService.this.getSystemService(Context.AUDIO_SERVICE);
    }

    @Override
    public void onDestroy() {
        super.onDestroy();

        removeNotification();
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            removeNotificationChannel();
        }

        if (mediaPlayer != null) {
            mediaPlayer.release();
        }
        mediaSessionCompat.release();

        castContext.getSessionManager().removeSessionManagerListener(castSessionManagerListener, CastSession.class);
        castContext.removeSessionTransferCallback(castSessionTransferCallback);
        removeAudioFocus();
        stopSelf();
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        Log.d(LOG_TAG, "onStartCommand called");

        if (intent.hasExtra(MainActivity.CURR_POSITION)) {
            currentPos = intent.getIntExtra(MainActivity.CURR_POSITION, 0);
        }

        MediaButtonReceiver.handleIntent(mediaSessionCompat, intent);

        return START_NOT_STICKY;
    }

    @Nullable
    @Override
    public BrowserRoot onGetRoot(@NonNull String clientPackageName, int clientUid, @Nullable Bundle rootHints) {
        return new BrowserRoot(MEDIA_ROOT_ID, null);
    }

    /** Add metadata below to populate your sample in the App. See README for instructions */
    @Override
    public void onLoadChildren(@NonNull String parentId, @NonNull Result<List<MediaBrowserCompat.MediaItem>> result) {
        mediaMetadata = new MediaMetadataCompat.Builder()
                .putString(MediaMetadataCompat.METADATA_KEY_MEDIA_ID, KEY_MEDIA_ID)
                .putString(MediaMetadataCompat.METADATA_KEY_TITLE, KEY_MEDIA_TITLE)
                .putString(MediaMetadataCompat.METADATA_KEY_DISPLAY_SUBTITLE, KEY_MEDIA_SUBTITLE)
                .putString(MediaMetadataCompat.METADATA_KEY_ARTIST, KEY_MEDIA_ARTIST)
                .putString(MediaMetadataCompat.METADATA_KEY_ALBUM, KEY_MEDIA_ALBUM)
                .putLong(MediaMetadataCompat.METADATA_KEY_DURATION,
                        TimeUnit.MILLISECONDS.convert(KEY_MEDIA_DURATION, TimeUnit.SECONDS))
                .putBitmap(MediaMetadataCompat.METADATA_KEY_ALBUM_ART,
                        BitmapFactory.decodeResource(getResources(), R.drawable.cast_album_art_placeholder))
                .putString(MediaMetadataCompat.METADATA_KEY_MEDIA_URI, KEY_MEDIA_URI)
                .build();
        mediaItems.add(new MediaBrowserCompat.MediaItem(mediaMetadata.getDescription(),
                MediaBrowserCompat.MediaItem.FLAG_PLAYABLE));
        result.sendResult(mediaItems);
    }

    private void checkMetadata() throws NullPointerException{
        if (KEY_MEDIA_TITLE == null || KEY_MEDIA_SUBTITLE == null || KEY_MEDIA_ARTIST == null ||
        KEY_MEDIA_ALBUM == null || KEY_MEDIA_ID == null) {
            Log.w(LOG_TAG, "One or more of Metadata values (TITLE, SUBTITLE, ARTIST, ALBUM or ID is null");
        }
        if (KEY_IMAGE_URI == null || KEY_MEDIA_URI == null) {
            throw new NullPointerException("One (or more) of the Metadata URI values is null and not set. \n" +
                    "Check and complete TODOs on top of AudioPlayerService.class mentioned in README.md");
        } else if (KEY_MEDIA_DURATION == 0) {
            String msg = "KEY_MEDIA_DURATION is 0. Seekbar will not appear in Notifications";
            showToast(msg);
        }
    }

    /** Audio Manager */
    @Override
    public void onAudioFocusChange(int focusChange) {
        Log.d(LOG_TAG, "onAudioFocusChanged");
        if (focusChange == AudioManager.AUDIOFOCUS_GAIN) {
            Log.d(LOG_TAG, "Focus Gained");
        } else if (focusChange == AudioManager.AUDIOFOCUS_LOSS) {
            Log.d(LOG_TAG, "Audio Focus Lost");
        } else if (focusChange == AudioManager.AUDIOFOCUS_LOSS_TRANSIENT) {
            Log.d(LOG_TAG, "Audio Focus Lost Transient");
        }
    }

    private void requestAudioFocus() {
        Log.d(LOG_TAG, "Speaker: "+audioManager.isSpeakerphoneOn());
        audioAttributes = new AudioAttributes.Builder()
                .setUsage(AudioAttributes.USAGE_MEDIA)
                .setContentType(AudioAttributes.CONTENT_TYPE_MUSIC)
                .build();
        audioFocusRequest = new AudioFocusRequest.Builder(AudioManager.AUDIOFOCUS_GAIN)
                .setAudioAttributes(audioAttributes)
                .setAcceptsDelayedFocusGain(false)
                .setOnAudioFocusChangeListener(this)
                .build();
        focusResult = audioManager.requestAudioFocus(audioFocusRequest);
        synchronized (focusLock) {
            if (focusResult == AudioManager.AUDIOFOCUS_REQUEST_GRANTED) {
                Log.d(LOG_TAG, "Focus Result: Granted");
                mediaPlayer.seekTo((int) currentPos);
            } else {
                Log.d(LOG_TAG, "Focus Result: "+focusResult);
            }
        }
    }

    private void removeAudioFocus() {
        audioManager.abandonAudioFocus(AudioPlayerService.this);
    }

    private void initMediaSession() {
        mediaSessionCompat = new MediaSessionCompat(getApplicationContext(), LOG_TAG);
        mediaSessionCompat.setFlags(
                MediaSessionCompat.FLAG_HANDLES_MEDIA_BUTTONS |
                        MediaSessionCompat.FLAG_HANDLES_TRANSPORT_CONTROLS);

        playbackStateBuilder = new PlaybackStateCompat.Builder()
                .setActions(PlaybackStateCompat.ACTION_PLAY |
                        PlaybackStateCompat.ACTION_PAUSE |
                        PlaybackStateCompat.ACTION_SEEK_TO |
                        PlaybackStateCompat.ACTION_STOP);
        playbackStateBuilder.setState(PlaybackStateCompat.STATE_NONE, 0, 1.0f);

        mediaSessionCompat.setPlaybackState(playbackStateBuilder.build());

        mediaSessionCallback = new MediaSessionCallback();

        mediaSessionCompat.setCallback(mediaSessionCallback);

        setSessionToken(mediaSessionCompat.getSessionToken());

        notificationManager = (NotificationManager) AudioPlayerService.this.getSystemService(Context.NOTIFICATION_SERVICE);

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            createNotificationChannel();
        }
        mediaPlayer = new MediaPlayer();
        playbackLocation = PlaybackLocation.LOCAL;
    }

    public void showToast(String msg) {
        Toast.makeText(this, msg, Toast.LENGTH_SHORT).show();
    }

    /** Notification */
    private void buildNotification(@NonNull @PlaybackStateCompat.State int state) {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            createNotificationChannel();
            MediaControllerCompat controller = mediaSessionCompat.getController();
            MediaMetadataCompat metadata = controller.getMetadata();
            MediaDescriptionCompat description = metadata.getDescription();

            MediaSession mediaSession = (MediaSession) mediaSessionCompat.getMediaSession();
            Notification.MediaStyle mediaStyle =
                    new Notification.MediaStyle()
                            .setMediaSession(mediaSession.getSessionToken())
                            .setShowActionsInCompactView(0,1,2);

            Notification.Builder builder = new Notification.Builder(this, CHANNEL_ID )
                    .setStyle(mediaStyle)
                    .setSmallIcon(R.mipmap.cast_hello_launcher);

            if (state == PlaybackStateCompat.STATE_PLAYING) {
                PendingIntent pauseIntent = MediaButtonReceiver.buildMediaButtonPendingIntent(
                        this,
                        PlaybackStateCompat.ACTION_PAUSE
                );

                Notification.Action pauseAction = new Notification.Action.Builder(
                        R.drawable.cast_ic_notification_pause,
                        "Pause",
                        pauseIntent
                ).build();

                builder.addAction(pauseAction);
            } else {
                PendingIntent playIntent = MediaButtonReceiver.buildMediaButtonPendingIntent(
                        getApplicationContext(),
                        PlaybackStateCompat.ACTION_PLAY
                );

                Notification.Action playAction = new Notification.Action.Builder(
                        R.drawable.cast_ic_notification_play,
                        "Play",
                        playIntent
                ).build();

                builder.addAction(playAction);
            }

            builder.setShowWhen(false);
            builder.setContentTitle(description.getTitle());
            builder.setLargeIcon(BitmapFactory.decodeResource(getResources(), R.mipmap.cast_hello_launcher));

            // Stop when notification is swiped away
            builder.setDeleteIntent(MediaButtonReceiver.buildMediaButtonPendingIntent(
                    getApplicationContext(),
                    PlaybackStateCompat.ACTION_STOP
            ));

            builder.setContentIntent(getContentIntent());

            builder.setVisibility(Notification.VISIBILITY_PUBLIC);

            Notification notification = builder.build();
            startForeground(NOTIFICATION_ID, notification);
        }
    }

    private PendingIntent getContentIntent() {
        Intent intent = new Intent(this.getApplicationContext(), MainActivity.class);
        return PendingIntent.getActivity(
                this.getApplicationContext(),
                0,
                intent,
                PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE
        );
    }

    // TODO - Implement in setDeleteIntent on Swipe
    private PendingIntent getDeleteIntent() {
        Intent intent = new Intent(this.getApplicationContext(), MainActivity.class);
        intent.putExtra(MainActivity.KILL_COMMAND, true);
        return PendingIntent.getActivity(
                this.getApplicationContext(),
                0,
                intent,
                PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_IMMUTABLE
        );
    }

    private void removeNotification() {
        notificationManager.cancelAll();
        Log.d(LOG_TAG, "removeNotification: Notification Removed");
    }

    public void createNotificationChannel() {
        if (notificationManager.getNotificationChannel(CHANNEL_ID) == null) {
            CharSequence name = "CastHelloAudio";
            String description = "CastHelloAudio Player";
            int importance = NotificationManager.IMPORTANCE_LOW;

            NotificationChannel channel = new NotificationChannel(CHANNEL_ID, name, importance);

            channel.setDescription(description);
            channel.enableLights(true);
            channel.setLightColor(Color.RED);
            channel.enableVibration(false);

            notificationManager.createNotificationChannel(channel);
            Log.d(LOG_TAG, "createNotificationChannel: New Channel Created");
        } else {
            Log.d(LOG_TAG, "createNotificationChannel: Existing Channel Re-Used");
        }
    }

    public void removeNotificationChannel() {
        if (notificationManager.getNotificationChannel(CHANNEL_ID) != null) {
            notificationManager.deleteNotificationChannel(CHANNEL_ID);
            Log.d(LOG_TAG, "removeNotificationChannel: Notification channel removed");
        } else {
            Log.d(LOG_TAG, "removeNotificationChannel: No existing channel to remove");
        }
    }

    public void initNotificationActions() {
        //TODO - Implement initNotificationActions after +/- skip30 actions have been implemented
    }

    public MediaInfo getMediaInfo() {
        MediaMetadata mediaMetadata = new MediaMetadata(MediaMetadata.MEDIA_TYPE_MUSIC_TRACK);
        mediaMetadata.putString(MediaMetadata.KEY_TITLE, KEY_MEDIA_TITLE);
        mediaMetadata.putString(MediaMetadata.KEY_ARTIST, KEY_MEDIA_ARTIST);
        mediaMetadata.putString(MediaMetadata.KEY_ALBUM_TITLE, KEY_MEDIA_ALBUM);
        mediaMetadata.putString(MediaMetadata.KEY_SUBTITLE, KEY_MEDIA_SUBTITLE);
        mediaMetadata.addImage(new WebImage(Uri.parse(KEY_IMAGE_URI)));

        MediaInfo mediaInfo = new MediaInfo.Builder(KEY_MEDIA_URI)
                .setStreamType(MediaInfo.STREAM_TYPE_BUFFERED)
                .setContentType("audio/wav")
                .setMetadata(mediaMetadata)
                .setStreamDuration(KEY_MEDIA_DURATION*1000)
                .build();

        return mediaInfo;
    }

    public void sendLoadRequest() {
        int currentPosition = 0;
        if (mediaPlayer != null) {
            currentPosition = mediaPlayer.getCurrentPosition();
            mediaPlayer.stop();
            mediaPlayer.release();
            mediaPlayer = null;
        }
        stopForeground(true);

        if (castSession != null) {
            remoteMediaClient = castSession.getRemoteMediaClient();
            remoteMediaClient.load(new MediaLoadRequestData.Builder()
                    .setMediaInfo(getMediaInfo())
                    .setAutoplay(true)
                    .setCurrentTime(currentPosition)
                    .build());
        }
    }

    private void initMediaPlayer() {
        if (mediaPlayer == null) {
            mediaPlayer = new MediaPlayer();
        }
        mediaPlayer.setAudioAttributes(
                new AudioAttributes.Builder()
                        .setContentType(AudioAttributes.CONTENT_TYPE_MUSIC)
                        .setUsage(AudioAttributes.USAGE_MEDIA)
                        .build()
        );
        try {
            mediaPlayer.setDataSource(KEY_MEDIA_URI);
            mediaPlayer.prepare();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void updatePlaybackState(int state) {
        playbackStateBuilder.setState(state, currentPos, 1.0f);
        mediaSessionCompat.setPlaybackState(playbackStateBuilder.build());
    }


    private void resumePlaybackRemoteToLocal(SessionState sessionState) {
        MediaLoadRequestData loadRequestData = sessionState.getLoadRequestData();
        currentPos = loadRequestData.getCurrentTime();

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            requestAudioFocus();
        }
        if (focusResult == AudioManager.AUDIOFOCUS_REQUEST_GRANTED) {
            mediaSessionCompat.getController().getTransportControls().play();
        }
    }

    /** Media Session  */
    private class MediaSessionCallback extends MediaSessionCompat.Callback {
        @Override
        public void onPrepare() {
            super.onPrepare();
            initMediaPlayer();
        }

        @Override
        public void onPlay() {
            super.onPlay();
            if (playbackLocation == PlaybackLocation.LOCAL) {
                if (!mediaPlayer.isPlaying()) {
                    if(mediaSessionCompat.getController().getMetadata() == null) {
                        mediaSessionCompat.setMetadata(mediaMetadata);
                    }
                    mediaSessionCompat.setActive(true);
                    currentPos = mediaPlayer.getCurrentPosition();
                    updatePlaybackState(PlaybackStateCompat.STATE_PLAYING);
                    mediaPlayer.start();
                    buildNotification(PlaybackStateCompat.STATE_PLAYING);
                } else {
                    showToast("Already playing");
                }
            } else {
                if (remoteMediaClient != null) {
                    if (!remoteMediaClient.hasMediaSession()) {
                        sendLoadRequest();
                    } else {
                        remoteMediaClient.play();
                    }
                }
            }
        }

        @Override
        public void onPause() {
            super.onPause();
            if (playbackLocation == PlaybackLocation.LOCAL) {
                mediaPlayer.pause();
                currentPos = mediaPlayer.getCurrentPosition();
                updatePlaybackState(PlaybackStateCompat.STATE_PAUSED);
                buildNotification(PlaybackStateCompat.STATE_PAUSED);
            } else {
                if (remoteMediaClient != null) {
                    remoteMediaClient.pause();
                }
            }
        }

        @Override
        public void onStop() {
            super.onStop();
            if (playbackLocation == PlaybackLocation.LOCAL) {
                mediaPlayer.stop();
                stopForeground(false);
                mediaPlayer.reset();
                initMediaPlayer();
                mediaSessionCompat.setActive(false);
                updatePlaybackState(PlaybackStateCompat.STATE_STOPPED);
                buildNotification(PlaybackStateCompat.STATE_STOPPED);
            } else {
                if (remoteMediaClient != null) {
                    remoteMediaClient.stop();
                }
            }
        }

        @Override
        public void onSeekTo(long pos) {
            super.onSeekTo(pos);
            int state = mediaSessionCompat.getController().getPlaybackState().getState();
            currentPos = pos;
            mediaPlayer.seekTo((int) pos);
            updatePlaybackState(state);
            buildNotification(state);
        }
    }

    /** Cast Session */
    private class CastSessionManagerListener implements SessionManagerListener<CastSession> {

        @Override
        public void onSessionEnded(@NonNull CastSession session, int error) {
            Log.d(LOG_TAG, "onSessionEnded called");
            playbackLocation = PlaybackLocation.LOCAL;
            castSession = null;
            String statusCodeString = CastStatusCodes.getStatusCodeString(error);
            Log.d(LOG_TAG, "**Ended** : "+statusCodeString);
            if (error == 2154) {
                stopForeground(true);
                stopSelf();
            }
        }

        @Override
        public void onSessionEnding(@NonNull CastSession session) {
            Log.d(LOG_TAG, "onSessionEnding called");
        }

        @Override
        public void onSessionResumeFailed(@NonNull CastSession session, int error) {
            Log.d(LOG_TAG, "onSessionResumeFailed called");
            playbackLocation = PlaybackLocation.LOCAL;
            castSession = null;
        }

        @Override
        public void onSessionResumed(@NonNull CastSession session, boolean wasSuspended) {
            Log.d(LOG_TAG, "onSessionResumed called");
            playbackLocation = PlaybackLocation.REMOTE;
            castSession = session;
        }

        @Override
        public void onSessionResuming(@NonNull CastSession session, @NonNull String sessionId) {
            Log.d(LOG_TAG, "onSessionResuming called");
        }

        @Override
        public void onSessionStartFailed(@NonNull CastSession session, int error) {
            Log.d(LOG_TAG, "onSessionStartFailed called");
            playbackLocation = PlaybackLocation.LOCAL;
            castSession = null;
        }

        @Override
        public void onSessionStarted(@NonNull CastSession session, @NonNull String sessionId) {
            Log.d(LOG_TAG, "onSessionStarted called");
            castSession = session;
            sendLoadRequest();
            playbackLocation = PlaybackLocation.REMOTE;
            Log.d(LOG_TAG, "Session id:"+sessionId);
        }

        @Override
        public void onSessionStarting(@NonNull CastSession session) {
            Log.d(LOG_TAG, "onSessionStarting called");
        }

        @Override
        public void onSessionSuspended(@NonNull CastSession session, int reason) {
            Log.d(LOG_TAG, "onSessionSuspended called");
            playbackLocation = PlaybackLocation.LOCAL;
        }
    }

    private class CastSessionTransferCallback extends SessionTransferCallback {
        public CastSessionTransferCallback() {
            super();
        }

        @Override
        public void onTransferFailed(int i, int i1) {
            Log.d(LOG_TAG, "onTransferFailed Called");

            if (i == TRANSFER_TYPE_FROM_REMOTE_TO_LOCAL) {
                if (i1 == TRANSFER_FAILED_REASON_STORE_SESSION_STATE) {
                    // Print logs
                    Log.e(LOG_TAG, "Remote-To-Local TransferFailedReason: Failed Store Session State");
                } else { // TRANSFER_FAILED_REASON_TRANSFER_TIMEOUT
                    // Print logs
                    Log.e(LOG_TAG, "Remote-To-Local TransferFailedReason: Transfer Timeout");
                }
            }
            stopSelf();
        }

        @Override
        public void onTransferred(int i, @NonNull SessionState sessionState) {
            Log.d(LOG_TAG, "onTransferred Called");
            if (i == TRANSFER_TYPE_FROM_REMOTE_TO_LOCAL) {
                resumePlaybackRemoteToLocal(sessionState);
            }
        }

        @Override
        public void onTransferring(int i) {
            Log.d(LOG_TAG, "onTransferring called");
            initMediaPlayer();
        }
    }
}
